@@ -12,6 +12,7 @@ from collections import OrderedDict
 from datetime import timedelta
 from enum import Enum
 from xml.dom import minidom
+from hashlib import sha256
 
 import requests
 from lxml import etree
@@ -20,12 +21,17 @@ from .functions import GetFunction, SetFunction
 from .models import (
     BandSetting,
     FilterAction,
+    FilterIpRange,
+    GuestNetworkEnabling,
+    GuestNetworkProperties,
     GuestNetworkSettings,
-    InterfaceGuestNetworkSettings,
     NatMode,
+    IPv6FilterRule,
+    IPv6FilterRuleProto,
     PortForward,
     Proto,
     RadioSettings,
+    RuleDir,
     SystemInfo,
     TimerMode,
 )
@@ -41,10 +47,21 @@ class Compal:
     Basic functionality for the router's API
     """
 
-    def __init__(self, router_ip, key=None, timeout=10):
+    def __init__(
+        self,
+        router_ip,
+        key=None,
+        send_token=True,
+        send_hash=False,
+        username="admin",
+        timeout=10,
+    ):
         self.router_ip = router_ip
+        self.send_token = send_token
+        self.send_hash = send_hash
+        self.username = username
         self.timeout = timeout
-        self.key = key
+        self.key = self.sanitize_key(key)
 
         self.session = requests.Session()
         # limit the number of redirects
@@ -65,14 +82,21 @@ class Compal:
         elif not self.initial_res.url.endswith("common_page/login.html"):
             LOGGER.error("Was not redirected to login page:" " concurrent session?")
 
+    def sanitize_key(self, key):
+        if key:
+            if self.send_hash:
+                key = sha256(key.encode("utf-8")).hexdigest()
+            else:
+                key = key[:31]
+        return key
+
     def initial_setup(self, new_key=None):
         """
         Replay the settings made during initial setup
         """
         LOGGER.info("Initial setup: english.")
 
-        if new_key:
-            self.key = new_key
+        self.key = self.sanitize_key(new_key)
 
         if not self.key:
             raise ValueError("No key/password availalbe")
@@ -85,7 +109,7 @@ class Compal:
         # Login or change password? Not sure.
         self.xml_setter(
             SetFunction.LOGIN,
-            OrderedDict([("Username", "admin"), ("Password", self.key)]),
+            OrderedDict([("Username", self.username), ("Password", self.key)]),
         )
         # Get current wifi settings (?)
         self.xml_getter(GetFunction.WIRELESSBASIC, {})
@@ -135,7 +159,8 @@ class Compal:
         (Which is a code smell)
         """
         data = OrderedDict()
-        data["token"] = self.session_token
+        if self.send_token:
+            data["token"] = self.session_token
 
         if "fun" in _data:
             data["fun"] = _data.pop("fun")
@@ -198,10 +223,14 @@ class Compal:
         Login. Allow this function to override the key.
         """
 
+        key = self.sanitize_key(key) if key else self.key
         res = self.xml_setter(
             SetFunction.LOGIN,
             OrderedDict(
-                [("Username", "admin"), ("Password", key if key else self.key)]
+                [
+                    ("Username", self.username),
+                    ("Password", key),
+                ]
             ),
         )
 
@@ -487,6 +516,7 @@ class Filters(object):
 
     def __init__(self, modem):
         self.modem = modem
+        self.parser = etree.XMLParser(recover=True)
 
     def set_parental_control(
         self,
@@ -567,15 +597,140 @@ class Filters(object):
 
         return self.modem.xml_setter(SetFunction.MACFILTER, {"data": data})
 
-    def set_ipv6_filter_rule(self):
+    def get_ipv6_filter_rules(self):
         """
-        To be integrated...
+        Get the list of IPv6 filter rules for incoming and outgoing traffic
         """
+
+        def r_int(rule, attr):
+            """
+            integer value for rule's child's text
+            """
+            return int(rule.find(attr).text)
+
+        def getFilterRules(direction):
+            """
+            Get the list of IPv6 filter rules for traffic in specified direction
+            """
+            res = self.modem.xml_getter(
+                GetFunction.IPV6FILTERING, OrderedDict([("rule", int(direction))])
+            )
+            xml = etree.fromstring(res.content, parser=self.parser)
+            for rule in xml.findall("instance"):
+                yield IPv6FilterRule(
+                    dir=direction,
+                    idd=r_int(rule, "idd"),
+                    src_addr=rule.find("src_addr").text,
+                    src_prefix=r_int(rule, "src_prefix"),
+                    dst_addr=rule.find("dst_addr").text,
+                    dst_prefix=r_int(rule, "dst_prefix"),
+                    src_sport=r_int(rule, "src_sport"),
+                    src_eport=r_int(rule, "src_eport"),
+                    dst_sport=r_int(rule, "dst_sport"),
+                    dst_eport=r_int(rule, "dst_eport"),
+                    protocol=IPv6FilterRuleProto(r_int(rule, "protocol")),
+                    allow=bool(r_int(rule, "allow")),
+                    enabled=bool(r_int(rule, "enabled")),
+                )
+
+        inRules = getFilterRules(RuleDir.incoming)
+        outRules = getFilterRules(RuleDir.outgoing)
+        return list(inRules), list(outRules)
+
+    def set_ipv6_filter_rule(self, rule):
+        """
+        Add a new IPv6 traffic filter rule
+        """
+
+        def getOrDefault(val, defaultValue):
+            return condGetOrDefault(RuleDir.incoming, val, defaultValue, defaultValue)
+
+        def condGetOrDefault(direction, val, defaultValueIn, defaultValueOut):
+            if direction == RuleDir.incoming:
+                return val if val is not None else defaultValueIn
+            else:
+                return val if val is not None else defaultValueOut
+
+        direction = getOrDefault(rule.dir, RuleDir.incoming)
+        src_prefix = getOrDefault(rule.src_prefix, 128)
+        src_addr = condGetOrDefault(direction, rule.src_addr, "::", None)
+        dst_prefix = getOrDefault(rule.dst_prefix, 128)
+        dst_addr = condGetOrDefault(direction, rule.dst_addr, None, "::")
         params = OrderedDict(
             [
-                ("act", ""),
-                ("dir", ""),
-                ("enabled", ""),
+                ("act", "2"),
+                ("dir", int(direction)),
+                ("enabled", int(getOrDefault(rule.enabled, True))),
+                (
+                    "allow_traffic",
+                    1 - int(condGetOrDefault(direction, rule.allow, True, False)),
+                ),  # 0 actually stands for allowing the traffic.. TODO introduce enum?
+                ("protocol", int(rule.protocol)),
+                ("src_addr", src_addr),
+                ("src_prefix", src_prefix),
+                ("dst_addr", dst_addr),
+                ("dst_prefix", dst_prefix),
+                ("ssport", condGetOrDefault(direction, rule.src_sport, "1", None)),
+                ("seport", condGetOrDefault(direction, rule.src_eport, "65535", None)),
+                ("dsport", condGetOrDefault(direction, rule.dst_sport, None, "1")),
+                ("deport", condGetOrDefault(direction, rule.dst_eport, None, "65535")),
+                ("del", ""),
+                ("idd", ""),
+                (
+                    "sIpRange",
+                    int(
+                        FilterIpRange.all
+                        if src_addr == "::"
+                        else FilterIpRange.range
+                        if src_prefix != 128
+                        else FilterIpRange.single
+                    ),
+                ),
+                (
+                    "dsIpRange",
+                    int(
+                        FilterIpRange.all
+                        if dst_addr == "::"
+                        else FilterIpRange.range
+                        if dst_prefix != 128
+                        else FilterIpRange.single
+                    ),
+                ),
+                ("PortRange", "2"),  # manual port selection
+                (
+                    "TMode",
+                    "0",
+                ),  # No timed rule
+                ("TRule", "0"),  # No timed rule
+            ]
+        )
+        return self.modem.xml_setter(SetFunction.IPV6_FILTER_RULE, params)
+
+    def update_ipv6_filter_rules(
+        self, ruleCount, disableIds={}, deleteIds={}, direction=RuleDir.incoming
+    ):
+        """
+        Update the existing filter set. I.e. disable or delete them.
+        """
+        if ruleCount == 0 or (len(disableIds) == 0 and len(deleteIds)):
+            pass
+
+        def genIddString():
+            return "*".join([str(i) for i in range(1, ruleCount + 1)])
+
+        def genBitfield(ids, matchStr, noMatchStr):
+            return "*".join(
+                [matchStr if i in ids else noMatchStr for i in range(1, ruleCount + 1)]
+            )
+
+        enableBitfield = genBitfield(disableIds, "0", "1")
+        deleteBitfield = genBitfield(deleteIds, "1", "0")
+
+        params = OrderedDict(
+            [
+                ("act", "1"),
+                ("dir", int(direction)),
+                ("enabled", enableBitfield),
                 ("allow_traffic", ""),
                 ("protocol", ""),
                 ("src_addr", ""),
@@ -586,17 +741,33 @@ class Filters(object):
                 ("seport", ""),
                 ("dsport", ""),
                 ("deport", ""),
-                ("del", ""),
-                ("idd", ""),
+                ("del", deleteBitfield),
+                ("idd", genIddString()),
                 ("sIpRange", ""),
                 ("dsIpRange", ""),
                 ("PortRange", ""),
-                ("TMode", ""),
-                ("TRule", ""),
+                ("TMode", "0"),  # No timed rule
+                ("TRule", "0"),  # No timed rule
             ]
         )
         return self.modem.xml_setter(SetFunction.IPV6_FILTER_RULE, params)
 
+    def delete_all_ipv6_filter_rules(self):
+        """
+        Delete all ipv6 filter rules
+        """
+        inRules, outRules = self.get_ipv6_filter_rules()
+        countInRules = len(inRules)
+        self.update_ipv6_filter_rules(
+            countInRules, {}, set(range(1, countInRules + 1)), RuleDir.incoming
+        )
+        countOutRules = len(outRules)
+        self.update_ipv6_filter_rules(
+            countOutRules, {}, set(range(1, countOutRules + 1)), RuleDir.outgoing
+        )
+
+        return inRules, outRules
+
     def set_filter_rule(self):
         """
         To be integrated...
@@ -1037,198 +1208,72 @@ class WifiGuestNetworkSettings(object):
         except (TypeError, ValueError):
             return val
 
-    @staticmethod
-    def __band_guest_networks(xml, band):
-        """
-        Get the wifi guest network settings for the given band (2g, 5g)
-        """
-        assert band in (
-            "2g",
-            "5g",
-        )
-
-        all_interfaces = list()
-        interfaces = xml.iter("Interface" + ("" if band == "2g" else "5G"))
-        for interface in interfaces:
-
-            def guest_xv(attr, coherce=True):
-                """
-                xml value for the given band
-                """
-                try:
-                    return WifiGuestNetworkSettings.__xml_value(
-                        interface, f"{attr}{band.upper()}", coherce
-                    )
-                except AttributeError:
-                    return WifiGuestNetworkSettings.__xml_value(
-                        interface, f"{attr}{band}", coherce
-                    )
-
-            all_interfaces.append(
-                InterfaceGuestNetworkSettings(
-                    radio=band,
-                    enable=guest_xv("Enable"),
-                    ssid=guest_xv("BSSID"),
-                    guest_mac=guest_xv("GuestMac"),
-                    hidden=guest_xv("HideNetwork"),
-                    re_key=guest_xv("GroupRekeyInterval"),
-                    security=guest_xv("SecurityMode"),
-                    pre_shared_key=guest_xv("PreSharedKey"),
-                    wpa_algorithm=guest_xv("WpaAlgorithm"),
-                )
-            )
-
-        return all_interfaces
-
     @property
     def wifi_guest_network_settings(self):
         """
-        Read the wifi guest network settings
+        Read the current wifi guest network settings for all wifi bands (2g and 5g).
         """
         xml = self.wifi_guest_network_settings_xml
-        guest_networks_2g = WifiGuestNetworkSettings.__band_guest_networks(xml, "2g")
-        guest_networks_5g = WifiGuestNetworkSettings.__band_guest_networks(xml, "5g")
+
+        # the compal modem returns 7 entries per band, the only used element is index 2
+        one_and_only_relevant_index = 2
+
+        interfaces = {
+            "2g": list(xml.iter("Interface"))[one_and_only_relevant_index],
+            "5g": list(xml.iter("Interface5G"))[one_and_only_relevant_index],
+        }
+
+        def guest_xv(band, attr, coherce=True):
+            """
+            xml value for the given band
+            """
+            try:
+                return WifiGuestNetworkSettings.__xml_value(
+                    interfaces[band], f"{attr}{band.upper()}", coherce
+                )
+            except AttributeError:
+                return WifiGuestNetworkSettings.__xml_value(
+                    interfaces[band], f"{attr}{band}", coherce
+                )
 
         return GuestNetworkSettings(
-            guest_networks_2g,
-            guest_networks_5g,
+            GuestNetworkEnabling(
+                guest_xv("2g", "Enable") == 1, guest_xv("2g", "GuestMac")
+            ),
+            GuestNetworkEnabling(
+                guest_xv("5g", "Enable") == 1, guest_xv("5g", "GuestMac")
+            ),
+            GuestNetworkProperties(
+                guest_xv("2g", "BSSID"),
+                guest_xv("2g", "HideNetwork"),
+                guest_xv("2g", "GroupRekeyInterval"),
+                guest_xv("2g", "SecurityMode"),
+                guest_xv("2g", "PreSharedKey"),
+                guest_xv("2g", "WpaAlgorithm"),
+            ),
         )
 
-    @staticmethod
-    def __compare_wifi_settings(old_settings, new_settings, interface_index, changes):
-        """
-        Compare two settings objects for changes and
-        return:
-            bln_changes: True if there were changes
-            changes: Python dict that contains the changes
-        """
-
-        def iterate_changes(old_settings, new_settings, changes):
-            for attr, old_value in old_settings.__dict__.items():
-                if isinstance(old_value, BandSetting):
-                    changes.update({attr: {}})
-                    iterate_changes(
-                        getattr(old_settings, attr),
-                        getattr(new_settings, attr),
-                        changes[attr],
-                    )
-                else:
-                    new_value = getattr(new_settings, attr)
-                    if old_value != new_value:
-                        changes.update({attr: f"{old_value} -> {new_value}"})
-                        bln_changes[0] = True
-
-        changes = {}
-        bln_changes = [False]
-        iterate_changes(
-            old_settings.guest_networks_2g[interface_index],
-            new_settings.guest_networks_2g[interface_index],
-            changes,
-        )
-        iterate_changes(
-            old_settings.guest_networks_5g[interface_index],
-            new_settings.guest_networks_5g[interface_index],
-            changes,
-        )
-        return bln_changes[0], changes
-
-    def __check_router_status(self, new_guest_network_settings, interface_index, debug):
-        """
-        Checks if (1) the new user settings were changed in the router (via checking
-        the getter fun:307 [WIRELESSGUESTNETWORK] in the router-xml-file), (2) prints out a
-        progress bar and (3) prints out if the process was successful.
-        """
-        router_settings = None
-
-        progress = 0
-
-        def print_progress_bar(progress, total, postfix=""):
-            progress = total if progress > total else progress
-            per_progress = int(progress / total * 100)
-            postfix = f"[{postfix}]" if postfix != "" else ""
-            print(
-                f"\r|{'█' * progress}{'-' * (total - progress)}| {per_progress}%, {progress}sec\t{postfix}",
-                end="",
-            )
-
-        total = 24
-        bln_changes = True
-        changes = ""
-        if debug:
-            print("\n--- WAITING FOR ROUTER TO UPDATE ---")
-        while progress < total and bln_changes is True:
-            time.sleep(3)
-            if debug:
-                print_progress_bar(progress, total, changes)
-            try:
-                router_settings = self.wifi_guest_network_settings
-                bln_changes, changes = self.__compare_wifi_settings(
-                    router_settings,
-                    new_guest_network_settings,
-                    interface_index,
-                    changes,
-                )
-            except Exception as e:
-                changes = str(e)
-        if debug:
-            if not bln_changes:
-                print("\n\n--- ROUTER SUCESSFULLY UPDATED ALL NEW WIFI SETTINGS! ---")
-            else:
-                print("\n\n--- CHANGES THAT DID NOT GET SET ---")
-                _, changes = self.__compare_wifi_settings(
-                    router_settings, new_guest_network_settings, interface_index
-                )
-                print(changes)
-        return bln_changes
-
-    def update_interface_guest_network_settings(
-        self, new_guest_network_settings, interface_index, debug=True
-    ):
+    def update_wifi_guest_network_settings(self, properties, enable):
         """
         Method for updating the wifi guest network settings. Uses fun:308.
+        The given properties are applied to all wifi bands (2g and 5g).
+        The enabling has only effect on wifi bands that are currently switched on.
+        Requires at least firmware CH7465LG-NCIP-6.15.30-1p3-1-NOSH.
         """
-        len_2g_list = len(new_guest_network_settings.guest_networks_2g)
-        len_5g_list = len(new_guest_network_settings.guest_networks_5g)
-        assert len_2g_list == len_5g_list
-        assert 0 <= interface_index < len_2g_list
-
-        def transform_interface(interface_settings):
-            out = []
-            out.extend(
-                [
-                    ("Interface", interface_index + 1),
-                    ("Enable", interface_settings.enable),
-                    ("Ssid", interface_settings.ssid),
-                    ("Hiden", interface_settings.hidden),
-                    ("Rekey", interface_settings.re_key),
-                    ("Security", interface_settings.security),
-                    ("PSkey", interface_settings.pre_shared_key),
-                    ("Wpaalg", interface_settings.wpa_algorithm),
-                ]
-            )
-
-            # Prefix 'wl', Postfix the band
-            return [(f"wl{k}{interface_settings.radio}", v) for (k, v) in out]
-
-        out_s = []  # change
-        for item_2g, item_5g in zip(
-            transform_interface(
-                new_guest_network_settings.guest_networks_2g[interface_index]
-            ),
-            transform_interface(
-                new_guest_network_settings.guest_networks_5g[interface_index]
-            ),
-        ):
-            out_s.append(item_2g)
-            out_s.append(item_5g)
-
-        out_settings = OrderedDict(out_s)
+        out_settings = OrderedDict(
+            [
+                ("wlEnable", 1 if enable else 2),
+                ("wlSsid", properties.ssid),
+                ("wlHiden", properties.hidden),
+                ("wlRekey", properties.re_key),
+                ("wlSecurity", properties.security),
+                ("wlPSkey", properties.pre_shared_key),
+                ("wlWpaalg", properties.wpa_algorithm),
+            ]
+        )
         self.modem.xml_setter(
             SetFunction.WIFI_GUEST_NETWORK_CONFIGURATION, out_settings
         )
-        return self.__check_router_status(
-            new_guest_network_settings, interface_index, debug
-        )
 
 
 class DHCPSettings:
@@ -1510,7 +1555,7 @@ class FuncScanner(object):
     def __init__(self, modem, pos, key):
         self.modem = modem
         self.current_pos = pos
-        self.key = key
+        self.key = modem.sanitize_key(key)
         self.last_login = -1
 
     @property
